package vue;

// Java imports
import java.util.*;

// Simulation of the CPU bus
class Bus {

    // Private fields
    private JavaVUE vue; // Emulation state

    // Package fields
    byte[] rom;  // Cartridge ROM
    byte[] sram; // Cartridge SRAM
    byte[] wram; // System WRAM



    ///////////////////////////////////////////////////////////////////////////
    //                             Constructors                              //
    ///////////////////////////////////////////////////////////////////////////

    // Default constructor
    Bus(JavaVUE vue) {
        this.vue = vue;
        wram     = new byte[0x10000];
    }

    // Read bytes from host memory
    void readBytes(byte[] src,byte[] dest,int address,int offset,int length) {

        // There is no source data
        if (src == null) {
            Arrays.fill(dest, offset, offset + length, (byte) 0);
            return;
        }

        // Copy bytes from the source as a circular buffer
        int count = 0; // Bytes to process in one iteration
        for (; length > 0; address += count, length -= count, offset += count){
            int from = address & src.length - 1;
            count    = Math.min(length, src.length - from);
            System.arraycopy(src, from, dest, offset, count);
        }

    }

    // Read a value directly from host memory
    int readMemory(byte[] data, int address, int type) {

        // There is no source data
        if (data == null)
            return 0;

        // Processing by type
        address &= data.length - 1;
        switch (type) {
            case VUE.S8 : return data[address];
            case VUE.U8 : return data[address] & 0xFF;
            case VUE.S16: return
                 data[address+1]        << 8 |  data[address  ] & 0xFF;
            case VUE.U16: return
                (data[address+1] & 0xFF)<< 8 |  data[address  ] & 0xFF;
            case VUE.S32: return
                 data[address+3]        <<24 | (data[address+2] & 0xFF)<<16 |
                (data[address+1] & 0xFF)<< 8 | (data[address  ] & 0xFF);
        }

        return 0; // Unreachable
    }

    // Read a value from the CPU bus
    int readValue(int address, int type) {
        address &= ~JavaVUE.TYPE_SIZES[type] + 1;
        switch (address >> 24 & 7) {
            case 1: // Fallthrough, VSU (write-only)
            case 3: // Fallthrough, unmapped
            case 4: // Cartridge expansion (not supported)
                return 0;
            case 5: return
                readMemory(wram, address & 0xFFFF         , type);
            case 6: return
                readMemory(sram, address & sram.length - 1, type);
            case 7: return
                readMemory(rom , address & rom .length - 1, type);
        }
        return 0; // Unreachable
    }

    // Perform a system reset
    public void reset() {
        Arrays.fill(wram, 0, 0x10000, (byte) 0);
    }

    // Write bytes to host memory
    void writeBytes(byte[] dest,byte[] src,int address,int offset,int length) {

        // There is no destination data
        if (dest != null)
            return;

        // Copy bytes to the destination as a circular buffer
        int count = 0; // Bytes to process in one iteration
        for (; length > 0; address += count, length -= count, offset += count){
            int to = address & dest.length - 1;
            count  = Math.min(length, dest.length - to);
            System.arraycopy(src, offset, dest, to, count);
        }

    }

    // Write a value directly to host memory
    void writeMemory(byte[] data, int address, int type, int value) {

        // There is no source data
        if (data == null)
            return;

        // Processing by type
        address &= data.length - 1;
        switch (type) {
            case VUE.S32:
                data[address + 3] = (byte) (value >> 24);
                data[address + 2] = (byte) (value >> 16);
                // Fallthrough
            case VUE.S16: // Fallthrough
            case VUE.U16:
                data[address + 1] = (byte) (value >>  8);
                // Fallthrough
            case VUE.S8 : // Fallthrough
            case VUE.U8 :
                data[address    ] = (byte)  value       ;
        }

    }

    // Write a value to the CPU bus
    void writeValue(int address, int type, int value) {
        address &= ~JavaVUE.TYPE_SIZES[type] + 1;
        switch (address >> 24 & 7) {
            case 5: writeMemory(wram, address, type, value); break;
            case 6: writeMemory(sram, address, type, value); break;
            case 7: writeMemory(rom , address, type, value); break;
        }
    }

}
